from controller.agent import AgentsController, AgentCard, AgentResult, BaseAgent
from pydantic import BaseModel
from dotenv import load_dotenv
load_dotenv()

import asyncio
import os
import re
import asyncio

from typing import Any

from langchain.agents import create_react_agent, AgentExecutor
from langchain.memory import ConversationBufferMemory
from langchain.tools import Tool
from langchain_openai import AzureChatOpenAI
from langchain.prompts import PromptTemplate
from collections import Counter
import logging
logging.basicConfig(
    level=logging.INFO,  # or logging.INFO
    format="%(asctime)s - %(levelname)s - %(message)s"
)

logger = logging.getLogger(__name__)

llm = AzureChatOpenAI(
    model=os.environ.get('MODEL_NAME'), 
    api_key=os.environ.get('AZURE_OPENAI_KEY'), #type:ignore
    azure_endpoint=os.environ.get('AZURE_OPENAI_BASE'),
    api_version=os.environ.get('AZURE_OPENAI_VERSION')
)
class MathAgent(BaseAgent):
    def __init__(self):
        super().__init__()

        def calculator(expression: str):
            """Calculate mathematical expressions safely"""
            try:
                allowed_chars = set('0123456789+-*/.() ')
                expression = expression.replace(" ", "").replace("'", "").replace('"', "")
                logger.info(f"evaluating expression : {expression}")
                if not all(c in allowed_chars for c in expression):
                    return "Error: Only basic mathematical operations allowed"
                
                result = eval(expression)
                return f"The result is: {result}"
            except Exception as e:
                return f"Error calculating: {str(e)}"

        def get_weather(city):
            """Get current weather for a city (mock function)"""
            weather_data = {
                "New York": "Sunny, 72°F",
                "London": "Cloudy, 15°C", 
                "Tokyo": "Rainy, 18°C",
                "Paris": "Partly cloudy, 20°C"
            }
            # return weather_data.get(city, f"Weather data not available for {city}")
            return "Sunny, 72°F"

        tools = [
            Tool(
                name="Calculator",
                func=calculator,
                description="Use this to perform mathematical calculations. Input should be a mathematical expression like '2+2' or '10*5'"
            ),
            Tool(
                name="Weather",
                func=get_weather,
                description="Get current weather for a city. Input should be the city name."
            ),
            Tool(
                name="supervisor_tool",
                description="Use this tool to communicate with supervisor",
                func=self.query_supervisor,
                coroutine=self.query_supervisor
            )
        ]

        prompt_template = """
        You are a helpful child agent that can use tools to answer questions and 
        you can use supervisor_tool to communcate with you supervisor if you need any other services or has doubt.

        You have access to the following tools:
        {tools}

        Use the following format:

        Question: the input question you must answer
        Thought: you should always think about what to do
        Action: the action to take, should be one of [{tool_names}]
        Action Input: the input to the action
        Observation: the result of the action
        ... (this Thought/Action/Action Input/Observation can repeat N times)
        Thought: I now know the final answer
        Final Answer: the final answer to the original input question

        Question: {input}
        Thought: {agent_scratchpad}"""

        prompt = PromptTemplate(
            template=prompt_template,
            input_variables=["input", "intermediate_steps", "tools", "tool_names", "agent_scratchpad"]
        )

        agent = create_react_agent(llm, tools, prompt)
        memory = ConversationBufferMemory(memory_key="chat_history", return_messages=True)
        self.agent = AgentExecutor(
            agent=agent,
            tools=tools,
            memory=memory,
            verbose=True,
            handle_parsing_errors=True,
            max_iterations=10
        )
    
    async def process(self, message: str) -> Any:
        await asyncio.sleep(1)
        print('myid ', self.uid)
        response = await self.agent.ainvoke({"input": message})
        
        return response['output']

class ContentAgent(BaseAgent):
    def __init__(self):
        super().__init__()

        def file_manager(operation):
            """
            Manage files - create, read, write, list files
            Format: "action:filename:content" or "action:filename" or "action"
            Actions: create, read, write, list, delete
            """
            try:
                parts = operation.split(":", 2)
                action = parts[0].lower()
                
                if action == "list":
                    files = [f for f in os.listdir(".") if f.endswith(('.txt', '.json', '.md'))]
                    return f"Available files: {', '.join(files) if files else 'No text files found'}"
                
                elif action == "create" and len(parts) >= 3:
                    filename, content = parts[1], parts[2]
                    with open(filename, 'w') as f:
                        f.write(content)
                    return f"File '{filename}' created successfully"
                
                elif action == "read" and len(parts) >= 2:
                    filename = parts[1]
                    if os.path.exists(filename):
                        with open(filename, 'r') as f:
                            content = f.read()
                        return f"Content of '{filename}':\n{content}"
                    else:
                        return f"File '{filename}' not found"
                
                elif action == "write" and len(parts) >= 3:
                    filename, content = parts[1], parts[2]
                    with open(filename, 'w') as f:
                        f.write(content)
                    return f"Content written to '{filename}'"
                
                elif action == "delete" and len(parts) >= 2:
                    filename = parts[1]
                    if os.path.exists(filename):
                        os.remove(filename)
                        return f"File '{filename}' deleted successfully"
                    else:
                        return f"File '{filename}' not found"
                
                else:
                    return "Invalid operation. Use format: 'action:filename:content' or 'action:filename' or 'list'"
            
            except Exception as e:
                return f"File operation error: {str(e)}"

        def text_analyzer(text):
            """
            Analyze text for various metrics and insights
            Returns word count, character count, most common words, readability info
            """
            try:
                if not text.strip():
                    return "No text provided for analysis"
                
                # Basic metrics
                char_count = len(text)
                char_count_no_spaces = len(text.replace(" ", ""))
                word_count = len(text.split())
                sentence_count = len([s for s in re.split(r'[.!?]+', text) if s.strip()])
                paragraph_count = len([p for p in text.split('\n\n') if p.strip()])
                
                # Word frequency analysis
                words = re.findall(r'\b\w+\b', text.lower())
                word_freq = Counter(words)
                most_common = word_freq.most_common(5)
                
                # Average metrics
                avg_words_per_sentence = word_count / sentence_count if sentence_count > 0 else 0
                avg_chars_per_word = char_count_no_spaces / word_count if word_count > 0 else 0
                
                # Simple readability estimate (based on sentence and word length)
                if avg_words_per_sentence > 20 or avg_chars_per_word > 6:
                    readability = "Complex"
                elif avg_words_per_sentence > 15 or avg_chars_per_word > 5:
                    readability = "Moderate"
                else:
                    readability = "Simple"
                
                analysis = f"""
        TEXT ANALYSIS RESULTS:
        ===================
        Characters: {char_count} (without spaces: {char_count_no_spaces})
        Words: {word_count}
        Sentences: {sentence_count}
        Paragraphs: {paragraph_count}
        Average words per sentence: {avg_words_per_sentence:.1f}
        Average characters per word: {avg_chars_per_word:.1f}
        Estimated readability: {readability}

        MOST COMMON WORDS:
        {chr(10).join([f"{word}: {count}" for word, count in most_common])}
                """
                
                return analysis.strip()
            
            except Exception as e:
                return f"Text analysis error: {str(e)}"

        tools = [
            Tool(
                name="FileManager",
                func=file_manager,
                description="""Manage files on the system. Use these formats:
                - 'list' - list available text files
                - 'create:filename.txt:content' - create new file with content
                - 'read:filename.txt' - read file content
                - 'write:filename.txt:content' - write content to file
                - 'delete:filename.txt' - delete file"""
            ),
            Tool(
                name="TextAnalyzer",
                func=text_analyzer,
                description="Analyze text for word count, character count, readability, most common words, and other metrics. Input should be the text to analyze.",
            ),
            Tool(
                name="supervisor_tool",
                description="Use this tool to communicate with supervisor, it will also provide weather informations",
                func=self.query_supervisor,
                coroutine=self.query_supervisor
            )
        ]

        prompt_template = """
        You are a helpful child agent that can use tools to answer questions and 
        you can use supervisor_tool to communcate with you supervisor if you need any other services or has doubt.

        You have access to the following tools:
        {tools}

        Use the following format:

        Question: the input question you must answer
        Thought: you should always think about what to do
        Action: the action to take, should be one of [{tool_names}]
        Action Input: the input to the action
        Observation: the result of the action
        ... (this Thought/Action/Action Input/Observation can repeat N times)
        Thought: I now know the final answer
        Final Answer: the final answer to the original input question

        Question: {input}
        Thought: {agent_scratchpad}"""

        prompt = PromptTemplate(
            template=prompt_template,
            input_variables=["input", "intermediate_steps", "tools", "tool_names", "agent_scratchpad"]
        )

        agent = create_react_agent(llm, tools, prompt)
        memory = ConversationBufferMemory(memory_key="chat_history", return_messages=True)
        self.agent = AgentExecutor(
            agent=agent,
            tools=tools,
            memory=memory,
            verbose=True,
            handle_parsing_errors=True,
            max_iterations=10
        )
    
    async def process(self, message: str) -> Any:
        await asyncio.sleep(1)
        print('myid ', self.uid)
        response = await self.agent.ainvoke({"input": message})
        
        return response['output']

class SimpleSuperVisor():
    def __init__(self):
        self.agent_controller = AgentsController(self, include_done_agent=True, llm=llm)

    def register_agents(self):
        card = AgentCard(name='mathematician', description='good at performing mathematical expressions', skills=['evaluating mathematical expressions', 'getting weather details'])
        self.agent_controller.register_agent(card, agent_class=MathAgent)

        card2 = AgentCard(name='content_writer', description='good at analyzing text informations and files managing', skills=['text analysis', 'files creating, editing, reading and deleting'], persistent=True)
        self.agent_controller.register_agent(card=card2, agent_class=ContentAgent)

    async def test(self):
        self.register_agents()
        res = self.agent_controller.registry.get_prompt_description()
        print(res)

        res = self.agent_controller.registry.create_agent_model()
        print(res.model_json_schema())

        # res = await self.agent_controller.execute_agent('content_writer', 'create a file named success.txt')
        # res = await self.agent_controller.execute_agent('content_writer', 'create a file named success2.txt')
        # print(res)
# Usage Example
# async def main():

    # Create supervisor
    # supervisor = SupervisorAgent(llm)
    
    # Register agents
    # supervisor.register_agent("AGENT1", "This agent can perform complex mathematical operations and has the ability to get weather details", MathAgent)
    # supervisor.register_agent("AGENT2", "This agent has the ability to manage files with operations including read, write, update or delete and this agent is also good in analysing text", ContentAgent, persistent=True)
    
    # try:
        # Test the system
#         result1 = await supervisor.solve_task("""
# This is a test to check your flow capabilities, follow these steps as said:
# 1. Ask AGENT2 to create a file named results.txt
# 2. Ask AGENT1 to calculate 1+2+3
# 3. Ask AGENT2 to store this result to results.txt
# """)
        # result1 = await supervisor.solve_task("calculate 1+2+3 and save result to a file")
        # print(f"Result 1: {result1}")
        
#         result2 = await supervisor.solve_task("""
# This is to test your capability, do the below step exactly:
# use agent1 to solve 1+5+9.
# ask agent2 to store this result and weather details of newyork.
# agent2 mostlikely ask you to provide weather details then use agent1 to get details and provide it.
# when agent2 successfull complete task you can end.
# """)
#         print(f"Result 2: {result2}")
        
        # result3 = await supervisor.solve_task("Complex math problem")
        # print(f"Result 3: {result3}")
        
    # except KeyboardInterrupt:
        # print("Exiting...!")
        
if __name__ == "__main__":
    # asyncio.run(main())
    s = SimpleSuperVisor()
    asyncio.run(s.test())


from agentcurie.supervisor import SupervisorAgent, AgentHook, FuncHook
from agentcurie.supervisor.message_manager import MessageManager
from agentcurie.controller import Controller, AgentsController, ToolsController, AgentCard, BaseAgent

__all__ = [
    'SupervisorAgent',
    'MessageManager',
    'Controller',
    'AgentsController',
    'ToolsController',
    'AgentHook',
    'FuncHook',
    'AgentCard',
    'BaseAgent'
]


# type: ignore
import asyncio
from inspect import iscoroutinefunction, signature
from typing import Any, Callable, Optional, Type, Literal

from pydantic import BaseModel, Field, create_model, RootModel
from .views import (
	ToolsRegistry,
	RegisteredTool,
)
from agentcurie.controller.tool.views import ToolModel

class Registry:
	"""Service for registering and managing tools"""

	def __init__(self, exclude_tools: list[str] = []):
		self.registry = ToolsRegistry()
		self.exclude_tools = exclude_tools

	def _create_param_model(self, function: Callable) -> Type[BaseModel]:
		"""Creates a Pydantic model from function signature"""
		sig = signature(function)
		params = {
			name: (param.annotation, ... if param.default == param.empty else param.default)
			for name, param in sig.parameters.items()
		}
		# TODO: make the types here work
		return create_model(
			f'{function.__name__}_parameters',
			__base__=ToolModel,
			**params,  # type: ignore
		)

	def tool(
		self,
		description: str,
		param_model: Optional[Type[BaseModel]] = None,
	):
		"""Decorator for registering tools"""

		def decorator(func: Callable):
			# Skip registration if tool is in exclude_tools
			if func.__name__ in self.exclude_tools:
				return func

			# Create param model from function if not provided
			actual_param_model = param_model or self._create_param_model(func)

			# Wrap sync functions to make them async
			if not iscoroutinefunction(func):

				async def async_wrapper(*args, **kwargs):
					return await asyncio.to_thread(func, *args, **kwargs)

				# Copy the signature and other metadata from the original function
				async_wrapper.__signature__ = signature(func)
				async_wrapper.__name__ = func.__name__
				async_wrapper.__annotations__ = func.__annotations__
				wrapped_func = async_wrapper
			else:
				wrapped_func = func

			tool = RegisteredTool(
				name=func.__name__,
				description=description,
				function=wrapped_func,
				param_model=actual_param_model,
			)
			self.registry.tools[func.__name__] = tool
			return func

		return decorator

	async def execute_tool(self, tool_name: str, params: dict) -> Any:
		"""Execute a registered tool"""
		if tool_name not in self.registry.tools:
			raise ValueError(f'tool {tool_name} not found')

		tool = self.registry.tools[tool_name]
		try:
			# Create the validated Pydantic model
			validated_params = tool.param_model(**params)

			# Check if the first parameter is a Pydantic model
			sig = signature(tool.function)
			parameters = list(sig.parameters.values())
			is_pydantic = parameters and issubclass(parameters[0].annotation, BaseModel)

			if is_pydantic:
				return await tool.function(validated_params)
			
			return await tool.function(**validated_params.model_dump())

		except Exception as e:
			raise RuntimeError(f'Error executing tool {tool_name}: {str(e)}') from e

	def create_tool_model(self, exclude_tools: list[str]=[]) -> type[ToolModel]:
		"""Creates a Union of individual tool models from registered tool,
		used by LLM APIs that support tool calling & enforce a schema.
		"""
		from typing import Union

		available_tools: dict[str, RegisteredTool] = {}
		for name, tool in self.registry.tools.items():
			if name in exclude_tools:
				continue

			available_tools[name] = tool

		# Create individual tool models for each tools
		individual_tool_models: list[Type[BaseModel]] = []

		# print("AVAILABLE TOOLS : ", " ,".join([key for (key, value) in available_tools.items()]))
		for name, tool in available_tools.items():
			# Create an individual model for each tool that contains only one field
			individual_model = create_model(
				f'{name.title().replace("_", "")}ToolModel',
				__base__=ToolModel,
				**{
					name: (
						tool.param_model,
						Field(..., description=tool.description),
					)
				},
			)
			individual_tool_models.append(individual_model)

		if not individual_tool_models:
			return create_model('EmptyToolModel', __base__=ToolModel)

		# Create proper Union type that maintains ToolModel interface
		if len(individual_tool_models) == 1:
			result_model = individual_tool_models[0]

		else:
			# Create a Union type using RootModel that properly delegates ToolModel methods
			union_type = Union[tuple(individual_tool_models)]

			class ToolModelUnion(RootModel[union_type]):
				"""Union of all available tool models that maintains ToolModel interface"""

				def model_dump(self, **kwargs):
					"""Delegate model_dump to the underlying tool model"""
					if hasattr(self.root, 'model_dump'):
						return self.root.model_dump(**kwargs) #type:ginore
					return super().model_dump(**kwargs)
				
				def get_type(self) -> Literal['tool']:
					return 'tool'

			# Set the name for better debugging
			ToolModelUnion.__name__ = 'ToolModel'
			ToolModelUnion.__qualname__ = 'ToolModel'

			result_model = ToolModelUnion

		return result_model  # type:ignore

	def get_prompt_description(self) -> str:
		"""Get a description of all tools for the prompt"""
		return self.registry.get_prompt_description()



import time

def get_first_key_param(model: dict):
    for key, value in model.items():
        key_name = key
        param = value
        break

    return key_name, param

def generate_unique_id():
    return int(f"{int(time.time() * 1000) % 1000000:06d}")


from __future__ import annotations

import json
from typing import List, Optional
from pydantic import BaseModel
from langchain_core.messages import (
	AIMessage,
	BaseMessage,
	HumanMessage,
	SystemMessage,
	ToolMessage,
)

from .views import MessageHistory, MessageMetadata
from agentcurie.supervisor.prompts import SystemPrompt
from agentcurie.controller import AgentOutput
from agentcurie.supervisor.utils import get_first_key_param

import logging
logger = logging.getLogger(__name__)


class MessageManager:
	def __init__(
		self,
		task: str|None = None,
		max_input_tokens: int = 128000,
		max_error_length: int = 400,
		estimated_characters_per_token: int = 3,
		tools_and_agents_description: str|None = None,
		system_prompt: SystemPrompt|None = None,
		message_context: Optional[str] = None,
	):
		self.max_input_tokens = max_input_tokens
		self.history = MessageHistory()
		self.task = task
		self.estimated_characters_per_token = estimated_characters_per_token
		self.max_error_length = max_error_length
		self.message_context = message_context
		self.IMG_TOKENS = 10

		system_prompt = system_prompt if system_prompt is not None else SystemPrompt()

		if tools_and_agents_description:
			system_message = system_prompt.get_system_message(tools_and_agents_description)

			self.system_prompt = system_message
			self._add_message_with_tokens(system_message)

		if self.message_context:
			context_message = HumanMessage(content=self.message_context)
			self._add_message_with_tokens(context_message)

		if task:
			task_message = self.task_instructions(task)
			self._add_message_with_tokens(task_message)

	@staticmethod
	def task_instructions(task: str) -> HumanMessage:
		content = f'Your task is: {task}.'
		return HumanMessage(content=content)

	@staticmethod
	def task_instructions_raw(task: str) -> str:
		return f'Your task is: {task}.'

	def add_new_task(self, new_task: str) -> None:
		self.task = new_task
		content = (
			f'Your task is: {new_task}.'
		)
		msg = HumanMessage(content=content)
		self._add_message_with_tokens(msg)

	# Guide the model through states, in future if needed
	# def add_state_message(
	# 	self,
	# 	result: Optional[List[ActionResult]] = None,
	# 	step_info: Optional[AgentStepInfo] = None,
	# ) -> None:
	# 	"""Add browser state as human message"""

	# 	# if keep in memory, add to directly to history and add state without result
	# 	if result:
	# 		for r in result:
	# 			if r.include_in_memory:
	# 				if r.extracted_content:
	# 					msg = HumanMessage(content='Action result: ' + str(r.extracted_content))
	# 					self._add_message_with_tokens(msg)
	# 				if r.error:
	# 					msg = HumanMessage(content='Action error: ' + str(r.error)[-self.max_error_length :])
	# 					self._add_message_with_tokens(msg)
	# 				result = None  # if result in history, we dont want to add it again

	# 	# otherwise add state message and result to next message (which will not stay in memory)
	# 	state_message = AgentMessagePrompt(
	# 		state,
	# 		result,
	# 		include_attributes=self.include_attributes,
	# 		max_error_length=self.max_error_length,
	# 		step_info=step_info,
	# 	).get_user_message()
	# 	self._add_message_with_tokens(state_message)

	def add_ai_message(self, message:str) -> None:
		self._add_message_with_tokens(
			AIMessage(content=message)
		)

	def add_human_message(self, message:str) -> None:
		self._add_message_with_tokens(
			HumanMessage(content=message)
		)

	def add_raw_tool_message(self, message: str) -> None:
		self._add_message_with_tokens(
			ToolMessage(
				content=message,
			)
		)

	def add_tool_message(self, message:str, name:str) -> None:
		self._add_message_with_tokens(
			ToolMessage(
				content=message,
				name=name,
				tool_call_id=name
			)
		)

	def pretty_print_messages(self) -> None:
		logger.info('------------------------------------------------------------------------------------------------')
		for message in self.history.messages[2:]:
			message = message.message
			if isinstance(message, ToolMessage):
				logger.info(f"TOOL({message.name}): {message.content}")
			elif isinstance(message, AIMessage):
				logger.info(f"AI: {message.content}, calling tools: {', '.join([m.get('name') for m in message.tool_calls])}")
			elif isinstance(message, HumanMessage):
				logger.info(f"HUMAN: {message.content}")
			else:
				logger.info(f"{message.content}")
		logger.info('------------------------------------------------------------------------------------------------')

	def get_key_param(self, model: dict):
		for key, value in model.items():
			key_name = key
			param = value
			break

		return key_name, param

	def format_agentoutput(self, model_output: AgentOutput):

		message = f"{model_output.evaluation_previous_goal}, {model_output.memory}, {model_output.next_goal}, "
		choice = model_output.get_choice()
		choice_dump = model_output.action.choice.model_dump()

		if choice == 'agent':
			agent_name, agent_message = get_first_key_param(choice_dump)
			message += f"For this i will call agent '{agent_name}' and assign it task '{agent_message}'\n"
		elif choice == 'tool':
			message += f"for this i will call tool : {choice_dump}"
		else:
			raise ValueError('Unknow choice')
		
		return message

	def add_choice(self, model_output: AgentOutput) -> None:
		choice = model_output.get_choice()
		choice_json = model_output.action.choice.model_dump()

		if choice == 'agent':
			agent_name, message = get_first_key_param(choice_json)
			tool_calls = [
				{
					'name': agent_name,
					'args': {'message':message},
					'id': agent_name,
					'type': 'tool_call',
				}
			]

			msg = AIMessage(
				content=self.format_agentoutput(model_output),
				tool_calls=tool_calls
			)
		elif choice == 'tool':
			key_name, param = get_first_key_param(choice_json)

			tool_calls = [
				{
					'name': key_name,
					'args': param,
					'id': key_name,
					'type': 'tool_call',
				}
			]

			msg = AIMessage(
				content=self.format_agentoutput(model_output),
				tool_calls=tool_calls
			)
		else:
			raise ValueError("Unknow choice")
		
		self._add_message_with_tokens(msg)

	def add_response(self, action: AgentOutput, result: str) -> None:
		key_name, param = get_first_key_param(action.action.choice.model_dump())
		tool_message = ToolMessage(
			content=result,
			tool_call_id=key_name,
		)
		self._add_message_with_tokens(tool_message)

	def get_messages(self) -> List[BaseMessage]:
		"""Get current message list, potentially trimmed to max tokens"""

		msg = [m.message for m in self.history.messages]
		# debug which messages are in history with token count # log

		total_input_tokens = 0
		logger.info(f'Messages in history: {len(self.history.messages)}:')
		for m in self.history.messages:
			total_input_tokens += m.metadata.input_tokens

		logger.info(f'Total input tokens: {total_input_tokens}')

		return msg
	
	def _add_message_with_tokens(self, message: BaseMessage) -> None:
		"""Add message with token count metadata"""
		token_count = self._count_tokens(message)
		metadata = MessageMetadata(input_tokens=token_count)
		self.history.add_message(message, metadata)

	def _count_tokens(self, message: BaseMessage) -> int:
		"""Count tokens in a message using the model's tokenizer"""
		tokens = 0
		if isinstance(message.content, list):
			for item in message.content:
				if 'image_url' in item:
					tokens += self.IMG_TOKENS
				elif isinstance(item, dict) and 'text' in item:
					tokens += self._count_text_tokens(item['text'])
		else:
			msg = message.content
			if hasattr(message, 'tool_calls'):
				msg += str(message.tool_calls)  # type: ignore
			tokens += self._count_text_tokens(msg)
		return tokens

	def _count_text_tokens(self, text: str) -> int:
		"""Count tokens in a text string"""
		tokens = len(text) // self.estimated_characters_per_token  # Rough estimate if no tokenizer available
		return tokens

	def cut_messages(self):
		"""Get current message list, potentially trimmed to max tokens"""
		diff = self.history.total_tokens - self.max_input_tokens
		if diff <= 0:
			return None

		msg = self.history.messages[-1]

		# if list with image remove image
		if isinstance(msg.message.content, list):
			text = ''
			for item in msg.message.content:
				if 'image_url' in item:
					msg.message.content.remove(item)
					diff -= self.IMG_TOKENS
					msg.metadata.input_tokens -= self.IMG_TOKENS
					self.history.total_tokens -= self.IMG_TOKENS
					logger.debug(
						f'Removed image with {self.IMG_TOKENS} tokens - total tokens now: {self.history.total_tokens}/{self.max_input_tokens}'
					)
				elif 'text' in item and isinstance(item, dict):
					text += item['text']
			msg.message.content = text
			self.history.messages[-1] = msg

		if diff <= 0:
			return None

		# if still over, remove text from state message proportionally to the number of tokens needed with buffer
		# Calculate the proportion of content to remove
		proportion_to_remove = diff / msg.metadata.input_tokens
		if proportion_to_remove > 0.99:
			raise ValueError(
				f'Max token limit reached - history is too long - reduce the system prompt or task. '
				f'proportion_to_remove: {proportion_to_remove}'
			)
		logger.debug(
			f'Removing {proportion_to_remove * 100:.2f}% of the last message  {proportion_to_remove * msg.metadata.input_tokens:.2f} / {msg.metadata.input_tokens:.2f} tokens)'
		)

		content = msg.message.content
		characters_to_remove = int(len(content) * proportion_to_remove)
		content = content[:-characters_to_remove]

		# remove tokens and old long message
		self.history.remove_message(index=-1)

		# new message with updated content
		msg = HumanMessage(content=content)
		self._add_message_with_tokens(msg)

		last_msg = self.history.messages[-1]

		logger.debug(
			f'Added message with {last_msg.metadata.input_tokens} tokens - total tokens now: {self.history.total_tokens}/{self.max_input_tokens} - total messages: {len(self.history.messages)}'
		)

	def convert_messages_for_non_function_calling_models(self, input_messages: list[BaseMessage]) -> list[BaseMessage]:
		"""Convert messages for non-function-calling models"""
		output_messages = []
		for message in input_messages:
			if isinstance(message, HumanMessage):
				output_messages.append(message)
			elif isinstance(message, SystemMessage):
				output_messages.append(message)
			elif isinstance(message, ToolMessage):
				output_messages.append(HumanMessage(content=message.content))
			elif isinstance(message, AIMessage):
				# check if tool_calls is a valid JSON object
				if message.tool_calls:
					tool_calls = json.dumps(message.tool_calls)
					output_messages.append(AIMessage(content=tool_calls))
				else:
					output_messages.append(message)
			else:
				raise ValueError(f'Unknown message type: {type(message)}')
		return output_messages

	def merge_successive_human_messages(self, messages: list[BaseMessage]) -> list[BaseMessage]:
		"""Some models like deepseek-reasoner dont allow multiple human messages in a row. This function merges them into one."""
		merged_messages:list[BaseMessage] = []
		streak = 0
		for message in messages:
			if isinstance(message, HumanMessage):
				streak += 1
				if streak > 1:
					merged_messages[-1].content += message.content #type:ignore
				else:
					merged_messages.append(message)
			else:
				merged_messages.append(message)
				streak = 0
		return merged_messages

	def extract_json_from_model_output(self, content: str) -> dict:
		"""Extract JSON from model output, handling both plain JSON and code-block-wrapped JSON."""
		try:
			# If content is wrapped in code blocks, extract just the JSON part
			if content.startswith('```'):
				# Find the JSON content between code blocks
				content = content.split('```')[1]
				# Remove language identifier if present (e.g., 'json\n')
				if '\n' in content:
					content = content.split('\n', 1)[1]
			# Parse the cleaned content
			return json.loads(content)
		except json.JSONDecodeError as e:
			logger.warning(f'Failed to parse model output: {str(e)}')
			raise ValueError('Could not parse response.')



import logging
import sys
import json
import os
from logging.handlers import RotatingFileHandler

def addLoggingLevel(levelName, levelNum, methodName=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.
    """
    if not methodName:
        methodName = levelName.lower()

    if hasattr(logging, levelName):
        raise AttributeError(f'{levelName} already defined in logging module')
    if hasattr(logging, methodName):
        raise AttributeError(f'{methodName} already defined in logging module')
    if hasattr(logging.getLoggerClass(), methodName):
        raise AttributeError(f'{methodName} already defined in logger class')

    def logForLevel(self, message, *args, **kwargs):
        if self.isEnabledFor(levelNum):
            self._log(levelNum, message, args, **kwargs)

    def logToRoot(message, *args, **kwargs):
        logging.log(levelNum, message, *args, **kwargs)

    logging.addLevelName(levelNum, levelName)
    setattr(logging, levelName, levelNum)
    setattr(logging.getLoggerClass(), methodName, logForLevel)
    setattr(logging, methodName, logToRoot)


def setup_logging(stream=None, log_level=None, force_setup=False):
    """
    Setup logging configuration for browser-use.
    """
    try:
        addLoggingLevel('RESULT', 35)
    except AttributeError:
        pass

    log_type = log_level

    if logging.getLogger().hasHandlers() and not force_setup:
        return logging.getLogger('gpa')

    root = logging.getLogger()
    root.handlers = []

    class BrowserUseFormatter(logging.Formatter):
        def format(self, record):
            return super().format(record)

    console = logging.StreamHandler(stream or sys.stdout)

    if log_type == 'result':
        console.setLevel('RESULT')
        console.setFormatter(BrowserUseFormatter('%(message)s'))
    else:
        console.setFormatter(BrowserUseFormatter('%(levelname)-8s [%(name)s] %(message)s'))

    root.addHandler(console)

    if log_type == 'result':
        root.setLevel('RESULT')
    elif log_type == 'debug':
        root.setLevel(logging.DEBUG)
    else:
        root.setLevel(logging.INFO)

    gpa_logger = logging.getLogger('gpa')
    gpa_logger.propagate = False
    gpa_logger.addHandler(console)
    gpa_logger.setLevel(root.level)

    logger = logging.getLogger('gpa')

    third_party_loggers = [
        'WDM',
        'httpx',
        'selenium',
        'playwright',
        'urllib3',
        'asyncio',
        'langsmith',
        'langsmith.client',
        'openai',
        'httpcore',
        'charset_normalizer',
        'anthropic._base_client',
        'PIL.PngImagePlugin',
        'trafilatura.htmlprocessing',
        'trafilatura',
        'groq',
        'portalocker',
        'portalocker.utils',
    ]

    for logger_name in third_party_loggers:
        third_party = logging.getLogger(logger_name)
        third_party.setLevel(logging.ERROR)
        third_party.propagate = False

    log_dir = "logs"
    os.makedirs(log_dir, exist_ok=True)
    log_file = os.path.join(log_dir, "gpa.log")

    class JSONFormatter(logging.Formatter):
        def format(self, record):
            log_data = {
                "timestamp": self.formatTime(record),
                "level": record.levelname,
                "module": record.module,
                "message": record.getMessage(),
            }
            if record.exc_info:
                log_data["exception"] = self.formatException(record.exc_info)
            return json.dumps(log_data)

    file_handler = RotatingFileHandler(log_file, maxBytes=5 * 1024 * 1024, backupCount=5)
    file_handler.setFormatter(JSONFormatter())

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s] %(name)s: %(message)s"))

    logger.addHandler(file_handler)
    logger.addHandler(console_handler)

    return logger


from .service import Controller as ToolsController
from .registery.service import Registry as ToolsRegistry
from .views import ToolModel, ToolResult, AgentOutput as ToolGuidingAgentOutput

__all__ = [
    'ToolsController',
    'ToolsRegistry',
    'ToolModel',
    'ToolResult',
    'ToolGuidingAgentOutput'
]


# AGENT CONTROLLER LOGIC
from pydantic import BaseModel
from typing import Any, TypeVar, Type, Optional
from .registery.service import Registry
from .views import AgentResult
from .registery.views import AgentCard, AgentModel, BaseAgent
from langchain_openai.chat_models.base import BaseChatOpenAI
from langchain_core.messages import SystemMessage, HumanMessage

T = TypeVar('T', bound=BaseModel)

class Controller():
	def __init__(
		self,
		supervisor,
		include_done_agent:bool = False,
		llm:Optional[BaseChatOpenAI] = None
	):
		self.supervisor = supervisor
		self.registry = Registry()

		if include_done_agent:
			if llm is None:
				raise ValueError("llm must needed to register doneAgent")
			class Summary(BaseModel):
				summary: str

			card = AgentCard(name="done_agent", description="Finish agents execution with summary to provide user", skills=['provides users comprehensive summary'])
			class Done(BaseAgent):
				def __init__(self):
					self.llm = llm
					
				async def process(self, message: str) -> Any:
					sys_message = SystemMessage(content="Your good at summarising the results and providing user a good insight")
					res:Summary = await self.llm.with_structured_output(Summary).ainvoke([sys_message, HumanMessage(message)]) #type:ignore

					return res.summary
				
			self.registry.register_agent(card=card, agent_class=Done)

	def register_agent(self, card: AgentCard, agent_class: Type[BaseAgent]):
		"""Decorator for registering custom agent

		@param description: Describe the LLM what the function does (better description == better function calling)
		"""
		return self.registry.register_agent(card, agent_class)
	
	async def act(self, agent: AgentModel) -> AgentResult:
		try:
			for agent_name, message in agent.model_dump(exclude_unset=True).items():

				if not self.registry.is_agent_exist(agent_name):
					raise ValueError(f"Unknow agent : {agent_name}")
				
				instance = self.registry.get_agent_instance(agent_name)
				instance.set_supervisor(self.supervisor)
				# if instance.supervisor is None:
					# instance.set_supervisor(self.supervisor)

				res = await instance.process(message)
				return AgentResult(content=res)
			return AgentResult()
		except Exception as e:
			raise e
			return AgentResult(error=str(e))



"""MCP (Model Context Protocol) client integration for gpa.

This module provides integration between external MCP servers and gpa registry.
MCP tools are dynamically discovered and registered as gpa tools.
"""

import asyncio
import logging
import time
from typing import Any, Union

from pydantic import BaseModel, Field, create_model

from controller import ToolResult
from controller.tool.registery.service import Registry
from controller.tool.service import Controller

logger = logging.getLogger(__name__)

# Import MCP SDK
from mcp import ClientSession, StdioServerParameters, types
from mcp.client.stdio import stdio_client

MCP_AVAILABLE = True


class MCPClient:
	"""Client for connecting to MCP servers and exposing their tools as gpa actions."""

	def __init__(
		self,
		server_name: str,
		command: str,
		args: list[str] | None = None,
		env: dict[str, str] | None = None,
	):
		"""Initialize MCP client.

		Args:
			server_name: Name of the MCP server (for logging and identification)
			command: Command to start the MCP server (e.g., "npx", "python")
			args: Arguments for the command (e.g., ["@playwright/mcp@latest"])
			env: Environment variables for the server process
		"""
		self.server_name = server_name
		self.command = command
		self.args = args or []
		self.env = env

		self.session: ClientSession | None = None
		self._stdio_task = None
		self._read_stream = None
		self._write_stream = None
		self._tools: dict[str, types.Tool] = {}
		self._registered_actions: set[str] = set()
		self._connected = False
		self._disconnect_event = asyncio.Event()

	async def connect(self) -> None:
		"""Connect to the MCP server and discover available tools."""
		if self._connected:
			logger.debug(f'Already connected to {self.server_name}')
			return

		start_time = time.time()
		error_msg = None

		try:
			logger.info(f"🔌 Connecting to MCP server '{self.server_name}': {self.command} {' '.join(self.args)}")

			# Create server parameters
			server_params = StdioServerParameters(command=self.command, args=self.args, env=self.env)

			# Start stdio client in background task
			self._stdio_task = asyncio.create_task(self._run_stdio_client(server_params))

			# Wait for connection to be established
			retries = 0
			max_retries = 100  # 10 second timeout (increased for parallel test execution)
			while not self._connected and retries < max_retries:
				await asyncio.sleep(0.1)
				retries += 1

			if not self._connected:
				error_msg = f"Failed to connect to MCP server '{self.server_name}' after {max_retries * 0.1} seconds"
				raise RuntimeError(error_msg)

			logger.info(f"📦 Discovered {len(self._tools)} tools from '{self.server_name}': {list(self._tools.keys())}")

		except Exception as e:
			error_msg = str(e)
			raise

	async def _run_stdio_client(self, server_params: StdioServerParameters):
		"""Run the stdio client connection in a background task."""
		try:
			async with stdio_client(server_params) as (read_stream, write_stream):
				self._read_stream = read_stream
				self._write_stream = write_stream

				# Create and initialize session
				async with ClientSession(read_stream, write_stream) as session:
					self.session = session

					# Initialize the connection
					await session.initialize()

					# Discover available tools
					tools_response = await session.list_tools()
					self._tools = {tool.name: tool for tool in tools_response.tools}

					# Mark as connected
					self._connected = True

					# Keep the connection alive until disconnect is called
					await self._disconnect_event.wait()

		except Exception as e:
			logger.error(f'MCP server connection error: {e}')
			self._connected = False
			raise
		finally:
			self._connected = False
			self.session = None

	async def disconnect(self) -> None:
		"""Disconnect from the MCP server."""
		if not self._connected:
			return

		start_time = time.time()
		error_msg = None

		try:
			logger.info(f"🔌 Disconnecting from MCP server '{self.server_name}'")

			# Signal disconnect
			self._connected = False
			self._disconnect_event.set()

			# Wait for stdio task to finish
			if self._stdio_task:
				try:
					await asyncio.wait_for(self._stdio_task, timeout=2.0)
				except TimeoutError:
					logger.warning(f"Timeout waiting for MCP server '{self.server_name}' to disconnect")
					self._stdio_task.cancel()
					try:
						await self._stdio_task
					except asyncio.CancelledError:
						pass

			self._tools.clear()
			self._registered_actions.clear()

		except Exception as e:
			logger.error(f'Error disconnecting from MCP server: {e}')

	async def register_to_controller(
		self,
		controller: Controller,
		tool_filter: list[str] | None = None,
		prefix: str | None = None,
	) -> None:
		"""Register MCP tools as actions in the browser-use controller.

		Args:
			controller: Browser-use controller to register actions to
			tool_filter: Optional list of tool names to register (None = all tools)
			prefix: Optional prefix to add to action names (e.g., "playwright_")
		"""
		if not self._connected:
			await self.connect()

		registry = controller.registry

		for tool_name, tool in self._tools.items():
			# Skip if not in filter
			if tool_filter and tool_name not in tool_filter:
				continue

			# Apply prefix if specified
			action_name = f'{prefix}{tool_name}' if prefix else tool_name

			# Skip if already registered
			if action_name in self._registered_actions:
				continue

			# Register the tool as an action
			self._register_tool_as_action(registry, action_name, tool)
			self._registered_actions.add(action_name)

		logger.info(f"✅ Registered {len(self._registered_actions)} MCP tools from '{self.server_name}' as browser-use actions")

	def _register_tool_as_action(self, registry: Registry, action_name: str, tool: Any) -> None:
		"""Register a single MCP tool as a browser-use action.

		Args:
			registry: Browser-use registry to register action to
			action_name: Name for the registered action
			tool: MCP Tool object with schema information
		"""
		# Parse tool parameters to create Pydantic model
		param_fields = {}

		if tool.inputSchema:
			# MCP tools use JSON Schema for parameters
			properties = tool.inputSchema.get('properties', {})
			required = set(tool.inputSchema.get('required', []))

			for param_name, param_schema in properties.items():
				# Convert JSON Schema type to Python type
				param_type = self._json_schema_to_python_type(param_schema)

				# Determine if field is required
				if param_name in required:
					default = ...  # Required field
				else:
					default = param_schema.get('default', None)

				# Add field with description if available
				field_kwargs = {}
				if 'description' in param_schema:
					field_kwargs['description'] = param_schema['description']

				param_fields[param_name] = (param_type, Field(default, **field_kwargs))

		# Create Pydantic model for the tool parameters
		if param_fields:
			param_model = create_model(f'{action_name}_Params', __base__=BaseModel, **param_fields)
		else:
			# No parameters - create empty model
			param_model = None

		# Create async wrapper function for the MCP tool
		# Need to define function with explicit parameters to satisfy registry validation
		if param_model:
			# Type 1: Function takes param model as first parameter
			async def mcp_action_wrapper(params: param_model) -> ToolResult:  # type: ignore[no-redef]
				"""Wrapper function that calls the MCP tool."""
				if not self.session or not self._connected:
					return ToolResult(error=f"MCP server '{self.server_name}' not connected", success=False)

				# Convert pydantic model to dict for MCP call
				tool_params = params.model_dump()

				logger.debug(f"🔧 Calling MCP tool '{tool.name}' with params: {tool_params}")

				start_time = time.time()
				error_msg = None

				try:
					# Call the MCP tool
					result = await self.session.call_tool(tool.name, tool_params)

					# Convert MCP result to ToolResult
					extracted_content = self._format_mcp_result(result)

					return ToolResult(
						content=extracted_content
					)

				except Exception as e:
					error_msg = f"MCP tool '{tool.name}' failed: {str(e)}"
					logger.error(error_msg)
					return ToolResult(error=error_msg, success=False)
		else:
			# No parameters - empty function signature
			async def mcp_action_wrapper() -> ToolResult:  # type: ignore[no-redef]
				"""Wrapper function that calls the MCP tool."""
				if not self.session or not self._connected:
					return ToolResult(error=f"MCP server '{self.server_name}' not connected", success=False)

				logger.debug(f"🔧 Calling MCP tool '{tool.name}' with no params")

				start_time = time.time()
				error_msg = None

				try:
					# Call the MCP tool with empty params
					result = await self.session.call_tool(tool.name, {})

					# Convert MCP result to ToolResult
					extracted_content = self._format_mcp_result(result)

					return ToolResult(
						content=extracted_content
					)

				except Exception as e:
					error_msg = f"MCP tool '{tool.name}' failed: {str(e)}"
					logger.error(error_msg)
					return ToolResult(error=error_msg, success=False)

		# Set function metadata for better debugging
		mcp_action_wrapper.__name__ = action_name
		mcp_action_wrapper.__qualname__ = f'mcp.{self.server_name}.{action_name}'

		# Register the action with browser-use
		description = tool.description or f'MCP tool from {self.server_name}: {tool.name}'

		# Use the registry's action decorator
		registry.action(description=description, param_model=param_model)(
			mcp_action_wrapper
		)

		logger.debug(f"✅ Registered MCP tool '{tool.name}' as action '{action_name}'")

	def _format_mcp_result(self, result: Any) -> str:
		"""Format MCP tool result into a string for ToolResult.

		Args:
			result: Raw result from MCP tool call

		Returns:
			Formatted string representation of the result
		"""
		# Handle different MCP result formats
		if hasattr(result, 'content'):
			# Structured content response
			if isinstance(result.content, list):
				# Multiple content items
				parts = []
				for item in result.content:
					if hasattr(item, 'text'):
						parts.append(item.text)
					elif hasattr(item, 'type') and item.type == 'text':
						parts.append(str(item))
					else:
						parts.append(str(item))
				return '\n'.join(parts)
			else:
				return str(result.content)
		elif isinstance(result, list):
			# List of content items
			parts = []
			for item in result:
				if hasattr(item, 'text'):
					parts.append(item.text)
				else:
					parts.append(str(item))
			return '\n'.join(parts)
		else:
			# Direct result or unknown format
			return str(result)

	def _json_schema_to_python_type(self, schema: dict) -> Any:
		"""Convert JSON Schema type to Python type.

		Args:
			schema: JSON Schema definition

		Returns:
			Python type corresponding to the schema
		"""
		json_type = schema.get('type')

		# Handle enums
		if 'enum' in schema:
			return str

		# Type mappings
		type_mapping = {
			'string': str,
			'number': float,
			'integer': int,
			'boolean': bool,
			'array': list,
			'object': dict,
			'null': type(None),
		}

		# Default fallback
		base_type = type_mapping.get(json_type, str)

		# Handle nullable
		if schema.get('nullable', False):
			base_type = Union[base_type, None]

		# Handle arrays
		if json_type == 'array':
			items_schema = schema.get('items')
			if items_schema and isinstance(items_schema, dict) and 'type' in items_schema:
				# 🟢 RECURSE to get element type!
				item_type = self._json_schema_to_python_type(items_schema)
				return list[item_type]
			else:
				# ⚠️ Invalid or missing items type — fallback to untyped list
				return list[Any]

		# Handle nested object
		if json_type == 'object' and 'properties' in schema:
			fields = {}
			required = set(schema.get('required', []))
			for prop_name, prop_schema in schema['properties'].items():
				prop_type = self._json_schema_to_python_type(prop_schema)
				default = ... if prop_name in required else prop_schema.get('default', None)
				field_kwargs = {}
				if 'description' in prop_schema:
					field_kwargs['description'] = prop_schema['description']
				fields[prop_name] = (prop_type, Field(default, **field_kwargs))
			return create_model("DynamicNested", __base__=BaseModel, **fields)

		return base_type

	async def __aenter__(self):
		"""Async context manager entry."""
		await self.connect()
		return self

	async def __aexit__(self, exc_type, exc_val, exc_tb):
		"""Async context manager exit."""
		await self.disconnect()



# Export agent controllers when implemented
from .tool import ToolsController, ToolsRegistry, ToolModel, ToolResult
from .agent import AgentsController, AgentCard, AgentResult, BaseAgent, SuperVisor
from .views import ChoiceModel, ChoiceResult, AgentOutput
from .service import Controller

__all__ = [
    'ToolsController',
    'ToolsRegistry',
    'ToolModel',
    'ToolResult',
    'AgentsController',
    'AgentCard',
    'AgentResult',
    'Controller',
    'ChoiceModel',
    'ChoiceResult',
    'BaseAgent',
    'AgentOutput',
    'SuperVisor'
]


from .service import Controller as AgentsController
from .registery.views import AgentCard, BaseAgent, SuperVisor
from .views import AgentResult, AgentOutput, AgentModel

__all__ = [
    'AgentsController',
    'AgentCard',
    'AgentResult',
    'BaseAgent',
    'AgentOutput',
    'AgentModel',
    'SuperVisor'
]


from .service import MessageManager
from .views import MessageHistory, ManagedMessage, MessageMetadata

__all__ = [
    "MessageManager",
    "MessageHistory",
    "ManagedMessage",
    "MessageMetadata"
]


import uuid
import asyncio
from typing import Dict, Any, Optional, Literal, Callable, Awaitable, TYPE_CHECKING
from pydantic import BaseModel, ConfigDict, Field, create_model
from dataclasses import dataclass, field
from enum import Enum
from abc import ABC, abstractmethod
from langchain_core.messages.base import BaseMessage

if TYPE_CHECKING:
    from supervisor import SupervisorAgent
    from controller import BaseAgent

# Event Types
class EventType(Enum):
    SOLVE_REQUEST_SUPERVISOR = "solve_request_supervisor"
    SOLVE_REQUEST_CHILD_AGENT = "solve_request_child_agent"
    QUERY_REQUEST = "query_request"
    CHILD_RESPONSE = "child_response"
    RESPONSE = "response"
    ERROR = "error"

@dataclass
class Event:
    event_type: EventType
    payload: Dict[str, Any]
    message: Optional[str] = None
    id: str = uuid.uuid4().hex
    task_id: Optional[str] = None
    agent_name: Optional[str] = None
    timestamp: float = field(default_factory=lambda: asyncio.get_event_loop().time())

class AgentStatus(Enum):
    IDLE = "idle"
    PROCESSING = "processing"
    WAITING_FOR_QUERY = "waiting_for_query"
    QUERY_PROCESSED = "query_processed"
    QUERY_INTERCEPTED = "query_intercepted"
    COMPLETED = "completed"
    ERROR = "error"

@dataclass
class TaskContext:
    message: str
    id: str = uuid.uuid4().hex
    status: AgentStatus = AgentStatus.IDLE
    result: Optional[Any] = None
    error: Optional[str] = None

@dataclass
class AgentContext:
    agent_name: str
    message: str
    id: str = uuid.uuid4().hex
    query: Optional[str] = None
    status: AgentStatus = AgentStatus.IDLE
    instance_id: str  = uuid.uuid4().hex
    result: Optional[str] = None
    error: Optional[str] = None

@dataclass
class QueryMessage:
    agent_name: str
    message: str
    id: str = uuid.uuid4().hex

class AgentResult(BaseModel):
    content:str|None = None
    error:str|None = None
    messages:list[BaseMessage] = []
    resources:dict = {}

class Hook(BaseModel):
    order: Literal['before', 'after']

class FuncHook(Hook):
    """Executes given func with supervisor"""
    func: Callable[['SupervisorAgent'], Awaitable]

class AgentHook(FuncHook):
    """Passes given agent as first param to func along with supervisor"""
    agent_name: str
    func: Callable[['BaseAgent','SupervisorAgent'], Awaitable]


#type:ignore
import os
from pydantic import BaseModel
from controller import Controller, ToolResult, AgentCard, BaseAgent
from controller.views import AgentOutput
from langchain_core.messages import SystemMessage, HumanMessage
from dotenv import load_dotenv
load_dotenv()

from langchain_openai import AzureChatOpenAI
llm = AzureChatOpenAI(
    model=os.environ.get('MODEL_NAME'), 
    api_key=os.environ.get('AZURE_OPENAI_KEY'), #type:ignore
    azure_endpoint=os.environ.get('AZURE_OPENAI_BASE'),
    api_version=os.environ.get('AZURE_OPENAI_VERSION')
)

class SimpleToolModel(BaseModel):
    name: str
    age: int

class SimpleToolModel2(BaseModel):
    bank: str
    pan: int

class SimpleAgent():
    def __init__(self) -> None:
        self.controller = Controller(supervisor=self, llm=llm)

    def test(self):
        @self.controller.tool("simple description")
        async def simple_tool(param: SimpleToolModel):
            return ToolResult(content='simple result')
        
        @self.controller.tool('simple description 2')
        async def simple_tool_2(param: SimpleToolModel2):
            return ToolResult(content="simple result 2")
        
        card = AgentCard(name='simple_agent', description='simple agent to manage files', skills=['file manager'])
        class SimpleChildAgent(BaseAgent):
            def __init__(self) -> None:
                print('Initialized child agent')

        card2 = AgentCard(name='simple_agent_2', description='simple agent to do math', skills=['mathematical operations'])
        class SimpleChildAgent2(BaseAgent):
            def __init__(self) -> None:
                print('Initialized child agent 2')

        self.controller.register_agent(agent_card=card, agent_class=SimpleChildAgent)
        self.controller.register_agent(agent_card=card2, agent_class=SimpleChildAgent2)

        sys_prompt = self.controller.get_prompt_description()

        choice_model = self.controller.create_choice_model()
        # print(choice_model.model_json_schema())

        agent_out_model = AgentOutput.type_with_custom_tools_and_agents(choice_model) #type:ignore

        # message = "Call the second tool with bank as sbi and pan as 1234"
        message = "Call a agent that can go math operations and ask it to do 1+1"
        res = llm.with_structured_output(agent_out_model, method="function_calling").invoke(input=[
            SystemMessage(content=sys_prompt),
            HumanMessage(message)
        ]) #type:ignore

        print(res, '\n\n', res.model_dump())

        print(res, '\n 2 : ', res.action.choice.get_type(), '\n 3 :', type(res.action))

        print(res.get_choice())
        # print(type(res), res.choice.get_type(), res.model_dump_json()) #type:ignore

if __name__ == "__main__":
    agent = SimpleAgent()
    agent.test()


from controller.agent.registery.views import AgentModel
from pydantic import RootModel

def unwrap_agent_model(choice: AgentModel | RootModel) -> AgentModel:
    if isinstance(choice, RootModel):
        return choice.root
    return choice


from agentcurie.controller import Controller, AgentOutput, ChoiceModel, BaseAgent, AgentCard, SuperVisor
from agentcurie.controller.agent import AgentModel
from agentcurie.controller.tool import ToolModel
from .message_manager import MessageManager
from .prompts import SystemPrompt
from .utils import get_first_key_param
from .views import AgentResult, AgentContext, AgentStatus, FuncHook, AgentHook
from langchain_openai import AzureChatOpenAI
from langchain_openai.chat_models.base import BaseChatOpenAI
from typing import Type, TypeVar
from pydantic import BaseModel
from typing import Dict, Any
from uuid import uuid4

import os
import asyncio

import logging
logger = logging.getLogger(__name__)

T = TypeVar('T', bound=BaseModel)

class SupervisorAgent(SuperVisor):
    def __init__(self, llm:BaseChatOpenAI, func_hooks:list[FuncHook] = [], agent_hooks:list[AgentHook] = []):
        self.llm = llm
        self.controller = Controller(supervisor=self, llm=llm)
        
        # func hooks - function which gets executed between steps also has access to whole supervisor
        self.func_hooks_before = [hook for hook in func_hooks if hook.order == 'before']
        self.func_hooks_after = [hook for hook in func_hooks if hook.order == 'after']

        # agent hooks - function which gets executed between steps also has access to whole supervisor and mentioned agents
        self.agent_hooks_before = [hook for hook in agent_hooks if hook.order == 'before']
        self.agent_hooks_after = [hook for hook in agent_hooks if hook.order == 'after']

        # updated at runtime
        # Registered tools and agents schema ofr model guiding
        self.AgentOutputModel:Type[ChoiceModel] = self.controller.create_choice_model()
        self.AgentOutput: Type[AgentOutput] = AgentOutput.type_with_custom_tools_and_agents(self.AgentOutputModel)

        # updated at runtime
        # without done model schemas to avoid task ending while another agent is waiting for response
        self.AgentOutputModelWithoutDone:Type[ChoiceModel] = self.controller.create_choice_model(exclude_tools=['done'])
        self.AgentOutputWithoutDone: Type[AgentOutput] = AgentOutput.type_with_custom_tools_and_agents(self.AgentOutputModelWithoutDone)

        self.message_manager = MessageManager()

        # contexts to store child agent tasks
        self.agent_tasks: Dict[str, AgentContext] = {}
        self.pending_queries = {}

    def register_agent(self, agent_card:AgentCard, agent_class:Type[BaseAgent]):
        self.controller.register_agent(agent_card=agent_card, agent_class=agent_class)

    def register_tool(self, description: str, **kwargs):
        return self.controller.tool(description, **kwargs)

    async def get_structured_response(self, messages, response_model:  Type[T]) -> T:
        """Get structured response using LangChain"""
        try:
            response = await self.llm.with_structured_output(response_model).ainvoke(messages)
            return response #type:ignore
        except Exception as e:
            raise e

    async def prepare_agent(self, task) -> None:
        # update system prompt
        logger.info("\033[33m ---- CONTEXT INITIALIZING STARTED ---- \033[0m")
        if len(self.message_manager.history.messages) == 0:
            logger.info('updating system prompt')
            system_prompt = SystemPrompt()
            system_message = system_prompt.get_system_message(self.controller.get_prompt_description())
            self.message_manager._add_message_with_tokens(system_message)
            # update model schemas
            logger.info('creating agent model')
            AgentOutputModel = self.controller.create_choice_model()
            self.AgentOutputModel = AgentOutputModel
            self.AgentOutput = AgentOutput.type_with_custom_tools_and_agents(AgentOutputModel) #type:ignore

            AgentOutputModelWithoutDone = self.controller.create_choice_model(exclude_tools=['done'])
            self.AgentOutputModelWithoutDone = AgentOutputModelWithoutDone
            self.AgentOutputWithoutDone = AgentOutput.type_with_custom_tools_and_agents(AgentOutputModelWithoutDone)

        # update task instructions
        logger.info('updating task instruction')
        self.message_manager.add_new_task(task)
        logger.info("\033[33m ---- CONTEXT INITIALIZING COMPLETED ---- \033[0m")

    async def _solve_query(self, message, agent_name) -> str:
        """Resolve query from child agents"""

        if agent_name not in self.agent_tasks:
            raise RuntimeError("Query from a unknow agent task")
        
        # update agent_task status so main logic can continue and process the query
        self.pending_queries[agent_name] = message
        self.agent_tasks[agent_name].query = message
        self.agent_tasks[agent_name].status = AgentStatus.WAITING_FOR_QUERY

        while self.agent_tasks[agent_name].status != AgentStatus.QUERY_PROCESSED:
            await asyncio.sleep(1)

        result = self.agent_tasks[agent_name].result
        assert isinstance(result, str)

        # reset state
        # del self.agent_tasks[agent_name]
        self.agent_tasks[agent_name].result = None
        self.agent_tasks[agent_name].query = None

        del self.pending_queries[agent_name]

        return result

    async def run_child_agent_with_context(self, agent_name, choice:AgentModel) -> None:
        "Assign and execute tasks to child agent with context status"
        try:
            result = await self.controller.execute_agent(choice)
            self.agent_tasks[agent_name].result = result.content
            self.agent_tasks[agent_name].status = AgentStatus.COMPLETED
        except Exception as e:
            logger.info(f"{agent_name} produced error {str(e)}")
            raise e

    async def solve(self, task: str):
        "Main entry point"
        try:
            # prepare agent
            await self.prepare_agent(task)

            step = 1
            while True:
                logger.info(f"Task : {task}")
                logger.info(f"🪜 Step : {step}")
                messages = self.message_manager.get_messages()

                # get response model, also avoiding task completion when there is pending queries
                response_model = self.AgentOutput if not self.pending_queries else self.AgentOutputWithoutDone
                logger.info(f"Using response model {response_model}")
                
                next_action:AgentOutput = await self.get_structured_response(messages, response_model)
                # logger.info(f"Resoponse from model : {next_action.model_dump()}")

                self.message_manager.add_choice(next_action)
                logger.info(f"\033[32m 🔍{next_action.evaluation_previous_goal} \n📝{next_action.memory} \n💡{next_action.next_goal} \033[0m")
                
                try:
                    choice = next_action.action.choice.root
                except Exception as e:
                    choice = next_action.action.choice

                if next_action.get_choice() == 'agent':
                    assert isinstance(choice, AgentModel)

                    agent_name, agent_task = get_first_key_param(choice.model_dump())
                    self.agent_tasks[agent_name] = AgentContext(message=agent_task, agent_name=agent_name)
                    

                    if agent_name in self.pending_queries:
                        # resolve pending queries from child agents
                        pending_query = self.pending_queries[agent_name]
                        self.agent_tasks[agent_name].result = agent_task
                        self.agent_tasks[agent_name].status = AgentStatus.QUERY_PROCESSED
                        logger.info(f"\033[32m Responded to pending query \n{agent_name}:{pending_query}? \nanswer:{agent_task} \033[0m")
                    else:
                        for hook in self.agent_hooks_before:
                            await hook.func(self.controller.agents_controller.registry.get_agent_instance(hook.agent_name), self)

                        logger.info(f"\033[32m Calling agent : {agent_name} - assinging task : '{agent_task}' \033[0m")
                        # run child agent as a background task to enable non-blocking executions
                        asyncio.create_task(self.run_child_agent_with_context(agent_name, choice))
                    
                    # wait for child agent to process and update status
                    while self.agent_tasks[agent_name].status not in (AgentStatus.COMPLETED, AgentStatus.ERROR, AgentStatus.WAITING_FOR_QUERY):
                        await asyncio.sleep(1)

                    res, status = self.agent_tasks[agent_name].result, self.agent_tasks[agent_name].status

                    if status == AgentStatus.COMPLETED or status == AgentStatus.ERROR:
                        message = f"Result from agent - {agent_name}: {res}"
                        logger.info(f"\033[36m {message} \033[0m")

                        self.message_manager.add_response(next_action, message)

                        for hook in self.agent_hooks_after:
                            await hook.func(self.controller.agents_controller.registry.get_agent_instance(hook.agent_name), self)
                    
                    if status == AgentStatus.WAITING_FOR_QUERY:
                        self.agent_tasks[agent_name].status = AgentStatus.QUERY_INTERCEPTED
                        message = f"While processing request, agent {agent_name} has a query: {self.agent_tasks[agent_name].query}"
                        logger.info(f"\033[36m {message} \033[0m")

                        self.message_manager.add_response(next_action, message)

                elif next_action.get_choice() == 'tool':
                    assert isinstance(choice, ToolModel)
                    tool_name, params = get_first_key_param(choice.model_dump())
                    logger.info(f"\033[32m Calling tool : {tool_name} - params: {params} \033[0m")

                    for hook in self.func_hooks_before:
                        await hook.func(self)

                    result = await self.controller.execute_tool(choice)
                    assert isinstance(result.content, str)

                    self.message_manager.add_response(next_action, result.content)

                    for hook in self.func_hooks_after:
                        await hook.func(self)

                    logger.info(f"\033[36m Result from tool - {tool_name} : {result.content} \033[0m")

                    if result.is_done:
                        logger.info(f"\n\nCompleted!, {result.content}")
                        logger.info(f"total tokens: {self.message_manager.history.total_tokens}")
                        break

                step += 1

            # post processing
            logger.info(f"Task completed : {result.content}")
            return AgentResult(content=result.content, messages=self.message_manager.get_messages())

        except Exception as e:
            # return AgentResult(error=str(e))
            raise e

if __name__ == "__main__":
    pass


from controller import ToolsController
from mcp_client.service import MCPClient
import asyncio


async def main() -> None:
    # Connect to an MCP server
    controller = ToolsController()
    mcp_client = MCPClient(
        server_name="my-server",
        command="npx",
        args=["-y", "@modelcontextprotocol/server-filesystem",r"C:\path\to\your\folder"]
    )

    # Register all MCP tools as browser-use actions
    await mcp_client.register_to_controller(controller)
    print(controller.registry.get_prompt_description())

if __name__ == "__main__":
    asyncio.run(main())


from typing import Callable, Dict, Type, Literal
from pydantic import BaseModel, ConfigDict

class ToolModel(BaseModel):
    """Base model for dynamically created tool models"""
    model_config = ConfigDict(arbitrary_types_allowed=True)
     
    def get_type(self) -> Literal['tool']:
        return "tool"
	
class RegisteredTool(BaseModel):
	"""Model for a registered action"""

	name: str
	description: str
	function: Callable
	param_model: Type[BaseModel]

	model_config = ConfigDict(arbitrary_types_allowed=True)

	def prompt_description(self) -> str:
		"""Get a description of the action for the prompt"""
		skip_keys = ['title']
		s = f'{self.description}: \n'
		s += '{' + str(self.name) + ': '
		s += str(
			{
				k: {sub_k: sub_v for sub_k, sub_v in v.items() if sub_k not in skip_keys}
				for k, v in self.param_model.schema()['properties'].items()
			}
		)
		s += '}'
		return s


class ToolsRegistry(BaseModel):
	"""Model representing the action registry"""

	tools: Dict[str, RegisteredTool] = {}

	def get_prompt_description(self) -> str:
		"""Get a description of all actions for the prompt"""
		return '\n'.join([tool.prompt_description() for tool in self.tools.values()])



# TOOLS CONTROLLER LOGIC
from pydantic import BaseModel
from typing import TypeVar
from .registery.service import Registry
from .views import ToolModel, ToolResult
from .tool_views import Done

T = TypeVar('T', bound=BaseModel)

class Controller():
	def __init__(
		self,
		exclude_tools: list[str] = [],
		include_done_tool:bool = True
	):
		self.exclude_tools = exclude_tools
		self.registry = Registry(exclude_tools)

		if include_done_tool:
			@self.registry.tool("Complete task - provide a summary of results for the user. Set success=True if task completed successfully, false otherwise. Text should be your response to the user summarizing results.", param_model=Done)
			async def done(params: Done) -> ToolResult:
				return ToolResult(content=params.summary, is_done=True)

	def tool(self, description: str, **kwargs):
		"""Decorator for registering custom tools

		@param description: Describe the LLM what the function does (better description == better function calling)
		"""
		return self.registry.tool(description, **kwargs)
	
	async def act(self, tool: ToolModel) -> ToolResult:
		"""Execute a tool"""
		try:
			for tool_name, params in tool.model_dump(exclude_unset=True).items():
				if params is not None:
					# remove highlights
					result = await self.registry.execute_tool(tool_name, params)
					if isinstance(result, str):
						return ToolResult(content=result)
					elif isinstance(result, ToolResult):
						return result
					elif result is None:
						return ToolResult()
					else:
						raise ValueError(f'Invalid tool result type: {type(result)} of {result}')
			return ToolResult()
		except Exception as e:
			raise e



# MAIN CONTROLLER LOGIC
from .tool import ToolsController, ToolModel
from .agent import AgentsController, AgentCard, BaseAgent, AgentModel
from .agent import SuperVisor
from .views import ChoiceModel, ChoiceResult
from typing import Type, Union
from pydantic import create_model, Field
from langchain_openai.chat_models.base import BaseChatOpenAI

import logging
logger = logging.getLogger(__name__)

class Controller():
    def __init__(self, supervisor:SuperVisor, llm:BaseChatOpenAI) -> None:
        """
            llm: The language model instance.
            supervisor: Supervisor agent which will get registered inside all other child agents.
        """
        self.agents_controller = AgentsController(supervisor=supervisor, llm=llm)
        self.tools_controller = ToolsController()

    def register_agent(self, agent_card:AgentCard, agent_class:Type[BaseAgent]):
        self.agents_controller.register_agent(card=agent_card, agent_class=agent_class)

    def tool(self, description: str, **kwargs):
        return self.tools_controller.tool(description, **kwargs)

    def create_choice_model(self, exclude_agents:list[str] = [], exclude_tools:list[str] = []) -> Type[ChoiceModel]:#type:ignore
        tools_model = self.tools_controller.registry.create_tool_model(exclude_tools)
        agents_model = self.agents_controller.registry.create_agent_model(exclude_agents)

        union_model = Union[(tools_model, agents_model)]

        model:ChoiceModel = create_model(
            'ChoiceModel',
            __base__=ChoiceModel,
            **{
                'choice':(#type:ignore
                    union_model,
                    Field(..., description="Tool or Agent")
                )
            }
        )

        return model#type:ignore

    def get_prompt_description(self) -> str:
        tools_description = self.tools_controller.registry.get_prompt_description()
        agents_description = self.agents_controller.registry.get_prompt_description()

        description = 'Available tools:'
        description += f'\n {tools_description}'
        description += f'\n Available agents'
        description += f'\n {agents_description}'

        return description

    async def execute_agent(self, choice:AgentModel) -> ChoiceResult:
        try:
            res = await self.agents_controller.act(choice)
            return ChoiceResult(content=res.content, is_done=res.is_done, success=res.success, error=res.error)
        except Exception as e:
            raise e
            return ChoiceResult(error=str(e))

    async def execute_tool(self, choice:ToolModel) -> ChoiceResult:
        try:
            logger.info(f'trying to execute tool : {choice}, {type(choice)}')
            res = await self.tools_controller.act(choice)
            return ChoiceResult(content=res.content, is_done=res.is_done, success=res.success, error=res.error)
        except Exception as e:
            logger.info(f"{str(e)}")
            raise e
            return ChoiceResult(error=str(e))

    async def execute_choise(self, choice:ChoiceModel) -> ChoiceResult:
        choosed_type = choice.get_type()

        try:
            if choosed_type == 'tool':
                assert isinstance(choice.choice.root, ToolModel)
                res = await self.tools_controller.act(choice.choice.root)
            elif choosed_type == 'agent':
                assert isinstance(choice.choice.root, AgentModel)
                res = await self.agents_controller.act(choice.choice.root)
            else:
                raise ValueError("Unknow choice, choice should be either a tool or an agent")
        
            return ChoiceResult(content=res.content, is_done=res.is_done, success=res.success, error=res.error)

        except Exception as e:
            return ChoiceResult(error=str(e))


from controller import ToolsController
from pydantic import BaseModel
import asyncio

class SimpleModel(BaseModel):
    query: str

async def main():
    controller = ToolsController()

    @controller.registry.tool("Test func1", param_model=SimpleModel)
    async def func1(params: SimpleModel):
        print('execute func1')

    @controller.registry.tool("Test func2", param_model=SimpleModel)
    async def func2(params: SimpleModel):
        print("executed func2", params)

    # get description of all tools with types specified
    res = controller.registry.get_prompt_description()
    print('desc ',res)

    # create union of tools models
    res = controller.registry.create_tool_model()
    print('param model ', res.model_json_schema())

    # execute tools
    res = await controller.registry.execute_tool('func1', {'query':'hi'})
    print(res)

if __name__ == "__main__":
    asyncio.run(main())
else:
    raise ValueError("tests cannot be imported")


from datetime import datetime
from typing import List, Optional

from langchain_core.messages import SystemMessage

class SystemPrompt:
	def __init__(self, current_date: datetime = datetime.now(), system_prompt: str|None = None):
		self.current_date = current_date
		self.system_prompt = system_prompt if system_prompt is not None else f"""
You are the Supervisor Agent — the coordinator and orchestrator of all other agents.
Your strength lies in reasoning, planning, and ensuring every child agent contributes effectively to achieve the user’s goal.

■ Objectives
1. Analyze the user's task and determine whether to:
   - Complete it directly, or
   - Delegate subtasks to suitable child agents.
2. Coordinate with child agents by issuing structured commands.
3. Resolve queries from child agents by:
   - Providing direct answers when possible, or
   - Consulting other agents to gather the needed information.
4. Ensure proper completion of all subtasks and handle pending queries before marking the main task as complete.
5. Finish the task by calling the `done` agent once all goals are satisfied and no unresolved queries remain.

■ Response Format
You must always respond in valid JSON using this exact structure:
{{
  "current_state": {{
    "evaluation_previous_goal": "Success | Failed | Unknown - Evaluate whether the previous action achieved its intended purpose based on the actual environment or webpage state (ignore the action result message). Mention briefly why it succeeded or failed, and note any unexpected outcomes (e.g., new suggestions appeared).",
    "memory": "Summarize what has been done so far and what should be remembered for the rest of the task.",
    "next_goal": "Describe what needs to be done next — the immediate actionable objective."
  }},
  "choice": {{
    "agent_name": "The name of the child agent whose service is needed next (or 'done' if the task is done)."
  }}
}}

■■ Supervision Rules
1. Task Decomposition:
   Break down complex user tasks into smaller, logical subtasks that can be executed sequentially or hierarchically by appropriate child agents.
2. Agent Selection:
   Select child agents based on their service descriptions.
   Example: If an agent provides web search capabilities, use it for information retrieval—not for decision-making.
3. Child agents have completely different context from you, so you should provide complete results to child agents.
   - If child_agent(X) or tool(X) provides you a result(R) you need to pass the complete result(R) to the next child agent(Y).
3. Query Handling:
   - If a child agent asks a query, try to answer directly.
   - If you lack sufficient information, delegate to another agent who can provide it.
   - Never assign a query back to the same agent that requested it.
4. Autonomy:
   If you can answer or complete the task directly (without involving other agents), do so.
5. Completion Condition:
   You cannot declare the task complete if there are any unresolved queries or pending subtasks.
   Once everything is resolved, invoke the `done` agent with the final output.
6. Consistency & Logging:
   Maintain logical consistency between memory, evaluation_previous_goal, and next_goal.
   Keep a mental record of context and past decisions to guide future steps.
7. Verification Before Completion:
   You must never assume a subtask was completed successfully without explicit confirmation from the agent responsible.
   Example: If a file needs to be written, wait for the FileAgent to confirm successful file creation before marking the task as done.
   Always base completion decisions on actual feedback, not reasoning or assumption.


Current date and time: {current_date}
1. RESPONSE FORMAT: You must ALWAYS respond with valid JSON in this exact format:
   {{
     "current_state": {{
       "evaluation_previous_goal": "Success|Failed|Unknown - Analyze the current elements and the image to check if the previous goals/actions are successful like intended by the task. Ignore the action result. The website is the ground truth. Also mention if something unexpected happened like new suggestions in an input field. Shortly state why/why not",
       "memory": "Description of what has been done and what you need to remember until the end of the task",
       "next_goal": "What needs to be done with the next actions"
     }},
     "action": {{
         action_name": {{
           // action-specific parameter
         }}
       }}
   }}

5. TASK COMPLETION:
   - Use the done tool as the last action as soon as the task is complete
   - Don't hallucinate actions
   - If the task requires specific information - make sure to include everything in the done function. This is what the user will see.
"""

	def get_system_message(self, description: str) -> SystemMessage:
		"""
		Get the system prompt for the agent.

		Returns:
		    str: Formatted system prompt
		"""

		AGENT_PROMPT = f"""
{self.system_prompt}

{description}

Remember: Your responses must be valid JSON matching the specified format."""
		return SystemMessage(content=AGENT_PROMPT)


from typing import Callable, Dict, Type, Optional, Literal
from pydantic import BaseModel, ConfigDict, Field, RootModel
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Type
import uuid

class AgentModel(BaseModel):
    """Base model for dynamically created agent models"""
    model_config = ConfigDict(arbitrary_types_allowed=True)
     
    def get_type(self) -> Literal['agent']:
        return "agent"
	
class AgentCard(BaseModel):
    """Defines agent information and skills"""
    name: str = Field(..., description="Your agent name, this will be used to uniquel access agent accross the system")
    description: str = Field(..., description="Describe what your agent does")
    skills: list[str] = Field(..., description="Descirbe set of skills your agent can perform")
    persistent: Optional[bool] = False

class SuperVisor(ABC):

    @abstractmethod
    async def _solve_query(self, message, agent_name) -> str:
        """Resolve query from child agents"""
        raise NotImplementedError('_solve_query must be overwrited')

class BaseAgent(ABC):
    def __init__(self):
        """
         BaseAgent from which All agents must inherit from.
         initialize your agent here and call the agent inside process function
        """
        self.agent_name:str|None = None # assigned in runtime
        self.supervisor:SuperVisor # assigned in runtime
        self.uid = uuid.uuid4().hex

        # tool = Tool(func=self.query_supervisor, desc="...")
        # self.agent = YourAgent(tools=[tool, ...])
        
    def set_supervisor(self, supervisor):
        self.supervisor = supervisor
    
    @abstractmethod
    async def process(self, message: str) -> Any:
        """Process the main task - should yield control back when needing to query"""
        # self.agent.invoke(message)
        raise RuntimeError("Not overrided")
    
    async def query_supervisor(self, query: str) -> str:
        """Tool to query the supervisor for help. If you stuck on any problem or need assistance, use this tool"""
        if not self.supervisor:
            raise RuntimeError("No supervisor set")

        if not self.agent_name:
            raise RuntimeError("Agent name not configured, SuperVisor Error")

        # message = QueryMessage(agent_id=self.agent_name, message=query)
        result = await self.supervisor._solve_query(query, self.agent_name)
        
        return result

class RegisteredAgent(BaseModel):
    card: AgentCard
    agent_class: Type[BaseAgent]

    def prompt_description(self):
        s = f'{self.card.name}: '
        s += f'{self.card.description}.'
        s += f' Skills: {', '.join(self.card.skills)}.'
        return s

class AgentsRegistry(BaseModel):
    """Model representing the action registry"""

    # name: card - for quick access
    agents: Dict[str, RegisteredAgent] = {}

    def get_prompt_description(self) -> str:
        """Get a description of all actions for the prompt"""
        return '\n'.join([agent.prompt_description() for agent in self.agents.values()])


from pydantic.json_schema import GenerateJsonSchema, JsonSchemaValue

def resolve_schema(schema: JsonSchemaValue) -> JsonSchemaValue:
    """Recursively resolve $ref references in a schema."""
    if not isinstance(schema, dict):
        return schema

    defs = schema.get("$defs", {})
    
    def _resolve(node):
        if isinstance(node, dict):
            if "$ref" in node:
                ref = node["$ref"]
                if ref.startswith("#/$defs/"):
                    name = ref.split("/")[-1]
                    return _resolve(defs.get(name, {}))
            return {k: _resolve(v) for k, v in node.items()}
        elif isinstance(node, list):
            return [_resolve(v) for v in node]
        return node

    return _resolve(schema) #type:ignore


from pydantic import BaseModel, ConfigDict, create_model, Field
from typing import Optional
from .registery.views import ToolModel

class ToolResult(BaseModel):
    """Result of executing an action"""		
    is_done: bool | None = False
    success: bool | None = True
    error: str | None = None
    content: str | None = None

# dynamically type guided model invoking schema for accurate tool params
class AgentOutput(BaseModel):
	model_config = ConfigDict(arbitrary_types_allowed=True, extra='forbid')

	evaluation_previous_goal: str
	memory: str
	next_goal: str
	choice: ToolModel = Field(
		...,
		description='tool to execute',
	)

	@staticmethod
	def type_with_custom_tools(tools_model: type[ToolModel]) -> type['AgentOutput']:
		"""Extend actions with custom actions"""

		model_ = create_model(
			'AgentOutput',
			__base__=AgentOutput,
			choice=(
				tools_model,  # type: ignore
				Field(..., description='tool to execute'),
			),
			__module__=AgentOutput.__module__,
		)
		model_.__doc__ = 'AgentOutput model with custom actions'
		return model_


# type: ignore
from typing import Dict, Type

from pydantic import BaseModel, Field, create_model, RootModel
from .views import (
	AgentsRegistry,
	RegisteredAgent,
	AgentCard
)
from .views import BaseAgent, RegisteredAgent
from .views import AgentModel

class Registry:
	"""Service for registering and managing agents"""

	def __init__(self):
		self.registry = AgentsRegistry()
		self.active_agents:Dict[str, BaseAgent] = {}

	def register_agent(self, card: AgentCard, agent_class: Type[BaseAgent],):
		"""function for registering agents"""
		self.registry.agents[card.name] = RegisteredAgent(card=card, agent_class=agent_class)

	def is_agent_exist(self, agent_name: str):
		if agent_name in self.registry.agents:
			return True
		
		return False

	def get_prompt_description(self) -> str:
		"""Get a description of all agents for the prompt"""
		return self.registry.get_prompt_description()
	
	def create_agent_instance(self, agent_name: str):
		if agent_name not in self.registry.agents:
			raise ValueError(f"Unknow agent type: {agent_name}")
		
		agent_class = self.registry.agents[agent_name].agent_class
		agent_instance = agent_class()
		agent_instance.agent_name = agent_name
		return agent_instance
	
	def get_agent_instance(self, agent_name) -> BaseAgent:
		if agent_name in self.active_agents:
			print("Getting from pre-initialized instances")
			return self.active_agents[agent_name]
		else:
			agent = self.create_agent_instance(agent_name)
			if self.registry.agents[agent_name].card.persistent:
				self.active_agents[agent_name] = agent

			return agent
		
	def delete_active_agent(self, agent_name) -> None:
		if agent_name in self.active_agents:
			del self.active_agents

		return
	
	def create_agent_model(self, exclude_agents:list[str] = []) -> Type[AgentModel]:
		from typing import Union
		
		individual_agent_models:list[Type[BaseModel]] = []

		# dynamic param or agent guided param initialization no added yet, will implement in future
		for agent_name, _ in self.registry.agents.items():
			if agent_name in exclude_agents:
				continue
			
			individual_agent_model = create_model(
				f'{agent_name.title().replace('_','')}AgentModel',
				__base__=AgentModel,
				**{
					agent_name:(
						str,
						Field(..., description=f'message you want to pass to {agent_name}')
					)
				}
			)
			individual_agent_models.append(individual_agent_model)

		if not individual_agent_models:
			return create_model('EmptyAgentModel', __base__=AgentModel)

		# Create proper Union type that maintains ToolModel interface
		if len(individual_agent_models) == 1:
			result_model = individual_agent_models[0]
		else:
			# Create a Union type using RootModel that properly delegates ToolModel methods
			union_type = Union[tuple(individual_agent_models)]

			class AgentModelUnion(RootModel[union_type]):
				"""Union of all available tool models that maintains ToolModel interface"""

				def model_dump(self, **kwargs):
					"""Delegate model_dump to the underlying tool model"""
					if hasattr(self.root, 'model_dump'):
						return self.root.model_dump(**kwargs) #type:ginore
					return super().model_dump(**kwargs)
				
				def get_type(self):
					return self.root.get_type()

			# Set the name for better debugging
			AgentModelUnion.__name__ = 'AgentModel'
			AgentModelUnion.__qualname__ = 'AgentModel'

			result_model = AgentModelUnion

		return result_model


from pydantic import BaseModel, ConfigDict, RootModel, Field, create_model
from .tool.registery.views import ToolModel
from .agent.registery.views import AgentModel
from typing import Union, Type, Literal

import logging
logger = logging.getLogger(__name__)

class ChoiceModel(BaseModel):
    choice: RootModel[Union[(ToolModel, AgentModel)]]
    model_config = ConfigDict(arbitrary_types_allowed=True)
    
    def get_type(self) -> Literal['tool', 'agent']:
        return self.root.get_type() #type:ignore
    
# dynamically type guided model invoking schema for accurate tool & agent params.
class AgentOutput(BaseModel):
	model_config = ConfigDict(arbitrary_types_allowed=True, extra='forbid')

	evaluation_previous_goal: str
	memory: str
	next_goal: str
	action: ChoiceModel = Field(
		...,
		description='tool or agent to execute',
	)

	@staticmethod
	def type_with_custom_tools_and_agents(choice_model: type[ChoiceModel]) -> type['AgentOutput']:
		"""Extend actions with custom actions"""

		model_ = create_model(
			'AgentOutput',
			__base__=AgentOutput,
			action=(
				choice_model,  # type: ignore
				Field(..., description='tool or agent to execute'),
			),
			__module__=AgentOutput.__module__,
		)
		model_.__doc__ = 'AgentOutput model with custom actions'
		return model_
      
	def get_choice(self) -> Literal['tool','agent']:
		try:
			try:
				choice = self.action.choice.root.get_type()
			except Exception as e:
				choice = self.action.choice.get_type() #type:ignore
				
			return choice
		except Exception as e:
			logger.info(e)
			raise e

class ChoiceResult(BaseModel):
	"""Result of executing an agent"""		
	is_done: bool | None = False
	success: bool | None = True
	error: str | None = None
	content: str | None = None
    
	def has_errors(self):
		if self.error is not None:
			return True
		
		return False


from __future__ import annotations

from typing import List

from langchain_core.messages import BaseMessage
from pydantic import BaseModel, Field


class MessageMetadata(BaseModel):
	"""Metadata for a message including token counts"""

	input_tokens: int = 0


class ManagedMessage(BaseModel):
	"""A message with its metadata"""

	message: BaseMessage
	metadata: MessageMetadata = Field(default_factory=MessageMetadata)


class MessageHistory(BaseModel):
	"""Container for message history with metadata"""

	messages: List[ManagedMessage] = Field(default_factory=list)
	total_tokens: int = 0

	def get_last_message(self):
		return self.messages[-1].message.content

	def add_message(self, message: BaseMessage, metadata: MessageMetadata) -> None:
		"""Add a message with metadata"""
		self.messages.append(ManagedMessage(message=message, metadata=metadata))
		self.total_tokens += metadata.input_tokens

	def remove_message(self, index: int = -1) -> None:
		"""Remove last message from history"""
		if self.messages:
			msg = self.messages.pop(index)
			self.total_tokens -= msg.metadata.input_tokens
